#!/usr/bin/perl -w
# ======================================================================                                                                                                  
# SBDiffPlain - Sentence Boundary Diff Tool for Unitex
# ====================================================================== 
# The script calculates recall, precision, and fscore value over all 
# sentences markers in unitex tagged key and response files. 
# The output is a csv file.
# ======================================================================  
# GPL copyright 2011 by Cristian Martinez <martinec|at|esiee.fr>
# ======================================================================                                                                                                 
use strict; 
use warnings;
# ======================================================================
use Text::CSV;
use Readonly;
use Tie::File;
use File::Temp qw/ tmpnam tempdir /;
use File::Copy;
use File::Basename;
use utf8;
# ======================================================================

# ======================================================================                 
# Constants 
# ======================================================================
#
Readonly my $RES_PTC 		=> "~~";	# Partially correct
Readonly my $RES_CRT 		=> "==";	# Correct
Readonly my $RES_MIS 		=> "+-";	# Missing
Readonly my $RES_FAP 		=> "-+";	# False positive
Readonly my $RES_MSM 		=> "<>";	# Mismatch

Readonly my $SYR_NWL 		=> "↵";		# New line
Readonly my $SYR_TAB 		=> "↹";		# Tab
Readonly my $SYR_SPC 		=> " ";		# Space
Readonly my $SYM_NWL1 	=> "\r\n";	# New line
Readonly my $SYM_NWL2 	=> "\n";	# New line
Readonly my $SYM_TAB 		=> "\t";	# Tab
Readonly my $SYM_SPC 		=> " ";		# Space

Readonly my $SYM_MSBD		=> "{S}";	# Main sentence boudary symbol
Readonly my $SYM_OSBD		=> "{\$}";	# Other sentence boudary symbol

Readonly my $LEN_MSBD		=> length($SYM_MSBD);
										# Length sentence boudary symbol 

Readonly my $CONTEXT_WIDTH	=> 32;		# Size of context for printing matches
Readonly my $EVAL_PREC		=> 5;		# Dec digits in Precn, Recall, Fscore

Readonly my $TMP_DIR		=> "sbdiff-XXXXXX";
Readonly my $TMP_FILE		=> "sbdiff-";

Readonly my $CVS_HEADER 	=> "No,Start,End,Context,Tp,Context,No,Start,End\n";
# ======================================================================                 
# Functions 
# ======================================================================
#
# escape string
sub escapestring($){
	my $str = shift;
	$str =~ s/$SYM_SPC/$SYR_SPC/g;	# Replace spaces
	$str =~ s/$SYM_TAB/$SYR_TAB/g;	# Replace tabs
	$str =~ s/$SYM_NWL1/$SYR_NWL/g;	# Replace new lines (\r\n)
	$str =~ s/$SYM_NWL2/$SYR_NWL/g;	# Replace new lines (\n)
	return $str;
}
# ======================================================================
# getcontext
sub getcontext($$$){
	my($left, $rigth, $sym_sep) = @_;
	
	$left 	= escapestring($left);
	$rigth = escapestring($rigth);
	
	my $head  	= substr $left,-$CONTEXT_WIDTH;
	my $tail  	= substr $rigth, 0,$CONTEXT_WIDTH;

	# ToDo: WTF!
	my $context = length($sym_sep)>0 ? $head.$sym_sep.$tail :
										   $head.$sym_sep.$tail;	
	return $context;
}
# ======================================================================
# printcvstr
sub printcsvstr($@){
	my ($csv,@fields) =  @_;
	
	if ($csv->combine (@fields)) {
	     my $string = $csv->string;
	     print "$string\n";
	}
	 else {
		 my $err = $csv->error_input;
	     print "combine () failed on argument: ", $err, "\n";
	}
}

# ======================================================================
# tmpfile
sub tmpfile($$){
	my $file	= shift;
	my $tempdir	= shift;
	
   (-e $file) or die "File doesn't exist: $file\n";
   
   my $tmpfn = File::Temp::tempnam( $tempdir, $TMP_FILE );
   copy($file,$tmpfn) or die "Copy of file failed: $file\n";
   
   return $tmpfn;
}

# ======================================================================                 
# Variables
# ======================================================================
#
my $count_ptc = 0;				# Partially correct count
my $count_crt = 0;				# Correct count
my $count_mis = 0;				# Missing count
my $count_fap = 0;				# False positive count
my $count_msm = 0;				# Mismatch count

# ======================================================================                 
# Main program  
# ====================================================================== 
#
if ($#ARGV != 1 ) {
	die "Usage: $0 key_document response_document\n"; 
}   

my $csv = Text::CSV->new ({
	binary => 1,always_quote => 1,}) 
	or die "Cannot use CSV: ".Text::CSV::Encoded->error_diag ();

my $tempdir = tempdir($TMP_DIR, TMPDIR => 1, CLEANUP => 1); 

my $keydocfn = tmpfile($ARGV[0],$tempdir);
my $resdocfn = tmpfile($ARGV[1],$tempdir);

use open ':encoding(utf8)';
binmode(STDOUT, ":utf8");

use Fcntl 'O_RDWR';
sysopen (KEYHANDLE, $keydocfn, O_RDWR) or die;
binmode(KEYHANDLE, ":encoding(utf8)");
tie my @keydoc, 'Tie::File', \*KEYHANDLE, recsep => $SYM_MSBD or 
				 die("Could not open temporaly document file: $keydocfn.\n");
				 
if($#keydoc==0){
	die ("Key document file without annotations: $keydocfn $#keydoc.\n");
}

use Fcntl 'O_RDWR';
sysopen (RESHANDLE, $resdocfn, O_RDWR) or die;
binmode(RESHANDLE, ":encoding(utf8)");
tie my @resdoc, 'Tie::File', \*RESHANDLE, recsep => $SYM_MSBD or
				die("Could not open temporaly$resdocfn document file.\n");
				
(tied @keydoc)->defer;
(tied @resdoc)->defer;

my ($keypos,$respos, $keytrim, $restrim,$keylen,$reslen);
$keypos = $respos = $keytrim = $restrim = $keylen= $reslen = 0;

my ($kstart, $kend, $rstart, $rend, $k, $r);
$kstart = $kend = $rstart = $rend = $k = $r = 0;

my $keystr;
my $resstr;
my $kcontext;
my $rcontext;
my @fields;

print $CVS_HEADER;

# Calcule differences between the two files
while (($k < $#keydoc)){
	
	$keystr= $keydoc[$k];
	$resstr= $resdoc[$r];
	$keylen= length($keydoc[$k]);
	$reslen= length($resdoc[$r]);
	
	if($keystr eq $resstr){
		
		$kstart	= $keypos;
		$kend  	= $kstart + $keylen + $keytrim;		   
		$kcontext= getcontext($keystr, $keydoc[$k+1],$SYM_MSBD);
		
		$rstart	= $respos;
		$rend  	= $rstart + $reslen + $restrim;
		$rcontext= getcontext($resstr, $resdoc[$r+1],$SYM_MSBD);
		
		@fields  = ($k+1,$kstart,$kend,$kcontext,$RES_CRT,$rcontext,$r+1,$rstart,$rend);
		printcsvstr($csv,@fields);

		$keypos+= $keylen + $LEN_MSBD + $keytrim; $keytrim = 0;
		$respos+= $reslen + $LEN_MSBD + $restrim; $restrim = 0;
		$r++; $k++;
		$count_crt++;
	}else{
		if($reslen < $keylen){	# False positive
			if(index($keystr, $resstr) == 0){  
				$rstart	= $respos;
				$rend  	= $rstart + $reslen + $restrim;		   
				$rcontext= getcontext($resstr, $resdoc[$r+1],$SYM_MSBD);
				
				$kstart	= $keypos;
				$kend  	= $kstart + $keylen + $keytrim;
				$keystr = $keydoc[$k] = substr $keystr, $reslen;
				$kcontext= getcontext($resstr, $keystr,"");
				
				@fields  = ($k+1,$kstart,$kend,$kcontext,$RES_FAP,$rcontext,$r+1,$rstart,$rend);
				
				$count_fap++;
				printcsvstr($csv,@fields);
			}else{										# ToDo: Partially or Mismatch
				
			}
			$keytrim+= $reslen;
			$respos+=  $reslen + $LEN_MSBD + $restrim; $restrim = 0;
			$r++;
		}else{											# Missing
			if(index($resstr,$keystr) == 0){
				$kstart	= $keypos;
				$kend  	= $kstart + $keylen + $keytrim;		   
				$kcontext= getcontext($keystr, $keydoc[$k+1],$SYM_MSBD);
				
				$rstart	= $respos;
				$rend  	= $rstart + $reslen + $restrim;
				$resstr = $resdoc[$r] = substr $resstr, $keylen;
				$rcontext= getcontext($keystr, $resstr,"");
				
				@fields  = ($k+1,$kstart,$kend,$kcontext,$RES_MIS,$rcontext,$r+1,$rstart,$rend);
				
				$count_mis++;	
				printcsvstr($csv,@fields);			
			}else{										# ToDo: Partially or Mismatch
				
			}
			$restrim+= $keylen;
			$keypos+=  $keylen + $LEN_MSBD + $keytrim; $keytrim = 0;
			$k++;
		}
	}	
}

# Calculate Precision and Recall and Fscore
my $recall 		= ($#keydoc == 0)? 0 : $count_crt / $#keydoc;
my $precision 	= ($#resdoc == 0)? 0 : $count_crt / $#resdoc;
my $fscore		= ($recall + $precision == 0) ? 0 : 
				  (2 * $recall * $precision) / ($recall + $precision);

# Print results
my $keybasename = basename($ARGV[0]);
my $resbasename = basename($ARGV[1]);

print  "\"\"\n";
print  "\"Key name\",\"$keybasename\"\n";
printf	"\"Key size\",\"%.0f\"\n", $#keydoc;
printf "\"Response name\",\"$resbasename\"\n", $#resdoc;
printf	"\"Response size\",\"%.0f\"\n", $#resdoc;
printf "\"Correct\",\"%.0f\"\n", $count_crt;
printf "\"Missing\",\"%.0f\"\n", $count_mis;
printf "\"False positives\",\"%.0f\"\n", $count_fap;
printf "\"Recall\",\"%.${\$EVAL_PREC}f\"\n", $recall; 
printf "\"Precision\",\"%.${\$EVAL_PREC}f\"\n", $precision; 
printf "\"Fscore\",\"%.${\$EVAL_PREC}f\"\n", $fscore;


# Clean close

(tied @keydoc)->flush;
(tied @resdoc)->flush;

untie @keydoc;            # all finished
untie @resdoc;            # all finished$

close(KEYHANDLE);
close(RESHANDLE);

$csv->eof or $csv->error_diag();

unlink($resdocfn) or print STDERR "Warning: Temporaly file was not deleted: $resdocfn\n";
unlink($keydocfn) or print STDERR "Warning: Temporaly file was not deleted: $keydocfn\n";


